#! /usr/bin/env python3

"""Write autoconf and CMake configuration for our C++ feature checks.

Produces configuration files for inclusion in the autoconf script and the
CMake configuration, respectively.  The config will ask the build systems to
try and compile each of the snippets of check code in `config-tests/`, and
where that succeeds, define the C++ macro of the same name as the base name
of the check snippet.
"""

from argparse import (
    ArgumentParser,
    Namespace,
)
import os.path
from pathlib import Path
from textwrap import dedent


# Name of the autoconf configuration script that we produce.  (Lives in the
# main source directory.)
AUTOCONF_CONFIG = "pqxx_cxx_feature_checks.ac"

# Name of the CMake config that we produce.  (Lives in the cmake directory.)
CMAKE_CONFIG = "pqxx_cxx_feature_checks.cmake"


def parse_args() -> Namespace:
    """Parse command line."""
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "--source",
        "-s",
        metavar="DIR",
        type=Path,
        default=Path(),
        help="Location of the source tree.  Defaults to current directory.",
    )
    return parser.parse_args()


def list_snippets(source: Path) -> list[str]:
    """Return list of C++ snippets."""
    return sorted(map(str, (source / "config-tests").glob("PQXX_*.cxx")))


# Comment header for either autoconf or CMake config we generate.
HEAD = dedent(
    """\
    # Configuration for feature checks. Generated by %(script)s.
    """
)


def compose_header() -> str:
    """Produce a config header."""
    return HEAD % {"script": os.path.basename(__file__)}


FOOT = "# End of config.\n"


# Template for an autoconf feature check.
AUTOCONF_TEMPLATE = dedent(
    """\
    AC_MSG_CHECKING([%(macro)s])
    %(macro)s=yes
    AC_COMPILE_IFELSE(
        [read_test(%(macro)s.cxx)],
        AC_DEFINE(
            [%(macro)s],
            1,
            [Define if this feature is available.]),
        %(macro)s=no)
    AC_MSG_RESULT($%(macro)s)
"""
)


def generate_autoconf(source: Path, snippets: list[str]) -> None:
    """Generate autoconf configuration."""
    dest = source / AUTOCONF_CONFIG
    with dest.open("w", encoding="ascii") as stream:
        stream.write(compose_header())
        for snippet in snippets:
            stream.write(AUTOCONF_TEMPLATE % {"macro": Path(snippet).stem})
        stream.write(FOOT)


# Template for a CMake feature check.
#
# Makes use of a custom function, try_compile.
CMAKE_TEMPLATE = dedent(
    """\
    try_compile(
        %(macro)s ${PROJECT_BINARY_DIR}
        SOURCES ${PROJECT_SOURCE_DIR}/config-tests/%(macro)s.cxx
    )
"""
)


def generate_cmake(source: Path, snippets: list[str]) -> None:
    """Generate CMake configuration."""
    dest = source / "cmake" / CMAKE_CONFIG
    with dest.open("w", encoding="ascii") as stream:
        stream.write(compose_header())
        for snippet in snippets:
            stream.write(CMAKE_TEMPLATE % {"macro": Path(snippet).stem})
        stream.write(FOOT)


def main(args: Namespace) -> None:
    """Main entry point."""
    snippets = list_snippets(args.source)
    generate_autoconf(args.source, snippets)
    generate_cmake(args.source, snippets)


if __name__ == "__main__":
    main(parse_args())



#! /usr/bin/env python3

"""Check that all source files are pure ASCII."""

from argparse import ArgumentParser
from pathlib import Path
import sys


def find_source_files(srcdir: Path) -> list[Path]:
    """Find source file paths in `srcdir`."""
    paths = []
    for dirpath, _, filenames in srcdir.walk():
        if dirpath.relative_to(srcdir) == Path("include") / "pqxx":
            # The header code in this directory is in files without suffixes.
            for filename in filenames:
                path = dirpath / filename
                if path.suffix in (".cxx", ".hxx") or "." not in filename:
                    paths.append(path)
        else:
            for filename in filenames:
                path = dirpath / filename
                if path.suffix in (".cxx", ".hxx"):
                    paths.append(path)
    return paths


def check_file(path: Path) -> list[int]:
    """Return any line numbers in file at `path` that are not pure ASCII.

    Actually we're not properly checking _lines,_ because there is no such
    notion in a binary file.  But we'll pretend we can split the contents by
    '\n', since it's mostly supposed to be ASCII (where this works).
    """
    lines = path.read_bytes().split(b"\n")
    bad_lines = []
    for n, data in enumerate(lines, start=1):
        # Why in blazes am I getting an int at the end of the file!?
        if isinstance(data, bytes):
            try:
                data.decode("ascii")
            except UnicodeDecodeError:
                bad_lines.append(n)
    return bad_lines


def parse_args():
    """Parse command line."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "-v",
        "--verbose",
        action="append_const",
        default=[],
        help="Print more detail.  Double this for even more.",
    )
    parser.add_argument(
        "-s",
        "--srcdir",
        default=Path(),
        type=Path,
        help="Source directory.  Defaults to current directory.",
    )
    return parser.parse_args()


def main() -> int:
    """Main entry point."""
    args = parse_args()

    files = sorted(find_source_files(args.srcdir))
    if len(args.verbose) > 1:
        print("\n".join(str(path) for path in files))

    ok = True
    for path in files:
        lines = check_file(path)
        if lines != []:
            ok = False
            print(f"*** Non-ASCII characters in {path}: ***", file=sys.stderr)
            for line in lines:
                print(f" - line {line}", file=sys.stderr)

    if ok:
        return 0
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())



#! /usr/bin/env python3

"""Extract relevant configuration macros from generated header.

We run `configure` or `cmake` to configure a build.  It spits out a config
header, but there are a lot of definitions in there that we don't actually
want.  Not just because they're not needed, but also because they pollute the
macro namespace.  It'd be fine if it was just inside libpqxx source files, but
the header affects client applications as well.

So, we rewrite the header (or its template) to contain only the macro
definitions with "PQXX" in the name.
"""

from argparse import (
    ArgumentParser,
    Namespace,
)
from pathlib import Path
import sys


# We prefix this to the header's text.  It suppresses a clang-tidy warning
# about there not being an ifdef/define pair to prevent multiple inclusion.
HEADER = "// NOLINT(llvm-header-guard)"


def parse_args() -> Namespace:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("infile", type=str, help="Input header.")
    parser.add_argument(
        "--outfile",
        "-o",
        type=str,
        help="Optional output header; defaults to same as input.",
    )
    return parser.parse_args()


def main() -> int:
    """Main entry point.  Returns exit code."""
    args = parse_args()
    infile = Path(args.infile)
    outfile = Path(args.outfile or args.infile)
    tmp = outfile.with_suffix(".tmp")
    with infile.open("r") as instream, tmp.open("w") as outstream:
        print(HEADER, file=outstream)
        for line in instream:
            if "PQXX" in line:
                print(line.strip(), file=outstream)

    # Now move the output file into place.  Doing this afterwards makes it
    # possible to use the same file for input and output.
    tmp.replace(outfile)
    return 0


if __name__ == "__main__":
    sys.exit(main())



#! /usr/bin/env python3

"""Check whether the compiler accepts various options.

Prints any of the flags that are accepted when used in combination with the
base command line and the previously accepted flags.
"""

from argparse import (
    ArgumentParser,
    Namespace,
)
from contextlib import nullcontext
from os import devnull
from pathlib import Path
import shlex
from subprocess import (
    DEVNULL,
    run,
)
import sys


EPILOG = """The flags file may contain comments, on lines starting with '#'
as the first non-whitespace character.  Any whitespace between flags other than
newlines is ignored, as are blank lines.
"""


def parse_args() -> Namespace:
    """Parse command-line options."""
    parser = ArgumentParser(description=__doc__, epilog=EPILOG)

    parser.add_argument(
        "-c",
        "--command",
        required=True,
        help="The basic compiler command line, e.g. 'c++ -std=c++20 -Werror'.",
    )
    parser.add_argument(
        "-f",
        "--flags",
        default="-",
        type=str,
        help="File with compiler flags to check, one per line; '-' for stdin.",
    )
    parser.add_argument(
        "-s",
        "--single-line",
        action="store_true",
        default=False,
        help="Print all command-line options on a single line.",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="-",
        type=str,
        help="Output file, or '-' for stdout (the default).",
    )

    return parser.parse_args()


def run_quietly(cmd: str) -> bool:
    """Run `cmd`, suppressing output; return whether it succeeds.

    **This will run `cmd` through shell interpretation.**
    """
    proc = run(cmd, stdout=DEVNULL, stderr=DEVNULL, shell=True, check=False)
    return proc.returncode == 0


def compiler_accepts(
    command: str, source: Path, flag: str, prev: str = ""
) -> bool:
    """Return whether the compiler seems to accept `flag`."""
    # This won't work with the default shell in Windows.  But then again who
    # would run the configure script there?
    src = shlex.quote(str(source))

    # I'd love to pass -c to prevent linking, but unfortunately gcc on Alpine
    # it accepts various sanitizer options for which it does not actually have
    # link-time support.
    return run_quietly(f"{command} {prev} {flag} {src} -o {devnull}")


def open_in(path: str):
    """Open file passed on command line for reading, unless it's '-'.

    If it's a dash (`-`), open standard input instead.
    """
    if path == "-":
        return nullcontext(sys.stdin)
    else:
        return Path(path).open()


def open_out(path: str):
    """Open file passed on command line for writing, unless it's '-'.

    If it's a dash (`-`), open standard output instead.
    """
    if path == "-":
        return nullcontext(sys.stdout)
    else:
        return Path(path).open("w")


def main() -> None:
    """Main entry point."""
    args = parse_args()
    good_flags: list[str] = []
    source = Path(__file__).parents[1] / "config-tests" / "minimal.cxx"
    with open_in(args.flags) as flags:
        for line in flags:
            flag = line.strip()
            if flag == "" or flag.startswith("#"):
                continue
            prev = " ".join(good_flags)
            if compiler_accepts(args.command, source, flag, prev=prev):
                good_flags.append(flag)

    sep = " " if args.single_line else "\n"

    with open_out(args.output) as output:
        print(sep.join(good_flags), file=output)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

"""Print libpqxx version information based on the source tree's VERSION file.

Version strings look like: "<major>.<minor>.<patch>".
"""

from argparse import ArgumentParser, Namespace
from os import getenv
from pathlib import Path
import sys


class Fail(Exception):
    """A well-handled exception.  Show message only, no traceback."""


def parse_args() -> Namespace:
    """Parse command-line arguments."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "--abi",
        "-a",
        action="store_true",
        help="Show ABI version only; omit patch.",
    )
    parser.add_argument(
        "--full",
        "-f",
        action="store_true",
        help="Show full version string (the default).",
    )
    parser.add_argument(
        "--major", "-M", action="store_true", help="Show major version."
    )
    parser.add_argument(
        "--minor", "-m", action="store_true", help="Show minor version."
    )
    parser.add_argument(
        "--patch", "-p", action="store_true", help="Show patch version."
    )
    parser.add_argument(
        "--patch-num",
        "-P",
        action="store_true",
        help="Show patch version if it's a number, or -1 otherwise.",
    )
    parser.add_argument(
        "--srcdir",
        "-s",
        type=Path,
        help="Source directory; defaults to '$srcdir' or '.'.",
    )
    args = parser.parse_args()
    actions = list(
        filter(
            None,
            [
                args.abi,
                args.full,
                args.major,
                args.minor,
                args.patch,
                args.patch_num,
            ],
        )
    )
    if len(actions) > 1:
        raise Fail(
            "Parse at most one of --abi, --full, --major, --minor, --patch, or --patch-num."
        )
    return args


def find_srcdir(args_srcdir: Path) -> Path:
    """Locate the root source directory."""
    return Path(args_srcdir or getenv("srcdir") or Path.cwd())


def read_version(srcdir: Path) -> str:
    """Read the VERSION file."""
    try:
        version = (srcdir / "VERSION").read_text().strip()
    except FileNotFoundError:
        raise Fail(f"Did not find VERSION file in {srcdir}.")
    if len(list(version.split("."))) != 3:
        raise Fail(f"Expected 3 components in version, found '{version}'.")
    return version


def patch_num(patch_str: str) -> int:
    """Return patch number as `int`, or -1 if it's not numeric."""
    if patch_str.isdigit():
        return int(patch_str)
    else:
        return -1


def work() -> None:
    """Do the actual work."""
    args = parse_args()
    version = read_version(find_srcdir(args.srcdir))
    if args.major:
        print(version.split(".")[0])
    elif args.minor:
        print(version.split(".")[1])
    elif args.patch:
        print(version.split(".")[2])
    elif args.patch_num:
        print(patch_num(version.split(".")[2]))
    elif args.abi:
        print(version.rsplit(".", 1)[0])
    else:
        # Default to full version.
        print(version)


def main() -> int:
    """Main entry point."""
    try:
        work()
    except Fail as e:
        print(e, file=sys.stderr)
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



#! /usr/bin/env python3

"""Generate autoconf/CMake checks for C++ feature check macros.

Produces feature checks for those features that we can detect based on just
the C++ feature check macros.

Reads the test libpqxx feature macro names, as well as the C++ feature check
macros that control them, from cxx_features.txt in the source tree.
"""

from argparse import (
    ArgumentParser,
    Namespace,
)
import os.path
from pathlib import Path
from textwrap import dedent
from typing import cast


# Name of the config file listing, one per line, the macros to test and set.
#
# Ignoring whitespace, a line can be:
# * empty, or
# * a comment starting with '#', or
# * a tuple of macro names.
#
# In that last case, the line must consist of a libpqxx feature macro name
# (which the configuration will either define or not define depening on whether
# there is support for the corresponding feature), followed by the name of the
# C++ feature test macro that we will need to check in order to determinue
# whether that feature is present.
CONFIG = "cxx_features.txt"


def parse_args() -> Namespace:
    """Parse command line."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "--source",
        "-s",
        metavar="DIR",
        type=Path,
        default=Path(),
        help="Location of the source tree.  Defaults to current directory.",
    )
    return parser.parse_args()


def read_config(config: Path) -> list[tuple[str, str]]:
    """Read config file, return as list of tuples.

    Each item is a tuple of a libpqxx feature macro, and the corresponding
    C++ feature check macro.
    """
    contents = config.read_text(encoding="ascii")
    lines = [line.strip() for line in contents.splitlines()]
    return [
        cast(tuple[str, str], tuple(line.split()))
        for line in lines
        if line != "" and not line.startswith("#")
    ]


# Template for a check snippet.
#
# Strictly speaking, this is unnecessarily verbose.  Why does it check first
# whether the macro is defined, and only then what its value is?  Because it's
# just conceivable that a user might have their compiler set to be extremely
# strict when it comes to checking undefined macros in a conditional.  The
# C++ standard says that in a conditional, an undefined macro shows up as if
# its value were zero.  But I can imagine some people wanting the compiler to
# warn them when they evaluate an undefined macro.
TEMPLATE = dedent(
    """\
    // Feature check for '%(pqxx_macro)s'.
    // Generated by %(script)s.
    #include <version>
    #if !defined(%(cxx_macro)s)
    #  error "No %(pqxx_macro)s: %(cxx_macro)s is not set."
    #endif
    #if !%(cxx_macro)s
    #  error "No %(pqxx_macro)s: %(cxx_macro)s is false."
    #endif

    int main() {}
    """
)


def write_check(
    source: Path, pqxx_macro: str, cxx_macro: str, base_params: dict
) -> None:
    """Write a piece of C++ for checking a feature macro."""
    params = dict(
        {
            "pqxx_macro": pqxx_macro,
            "cxx_macro": cxx_macro,
        },
        **base_params,
    )

    check = source / "config-tests" / f"{pqxx_macro}.cxx"
    check.write_text(TEMPLATE % params, encoding="ascii")


def main(args: Namespace) -> None:
    """Main entry point."""
    items = read_config(args.source / "cxx_features.txt")
    base_params = {
        "script": os.path.basename(__file__),
    }
    for pqxx_macro, cxx_macro in items:
        write_check(args.source, pqxx_macro, cxx_macro, base_params)


if __name__ == "__main__":
    main(parse_args())



#! /usr/bin/env python3

"""M4-quote text, for use as a literal in configure.ac.

Produces M4 "code" which evaluates to the input text.

It's not easy to read plain text from an input file in M4, without having it
expanded as M4.  Sometimes all we want is literal text!
"""
from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
)

from argparse import ArgumentParser
from sys import (
    stdin,
    stdout,
)


def parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "--open", "-a", default="[[", help="Current open-quote symbol."
    )
    parser.add_argument(
        "--close", "-b", default="]]", help="Current close-quote symbol."
    )
    parser.add_argument(
        "--input", "-i", default="-", help="Input file, or '-' for stdin."
    )
    parser.add_argument(
        "--output", "-o", default="-", help="Output file, or '-' for stdout."
    )
    return parser.parse_args()


def open_input(in_file):
    if in_file == "-":
        return stdin
    else:
        return open(in_file)


def open_output(out_file):
    if out_file == "-":
        return stdout
    else:
        return open(out_file, "w")


def escape(line):
    return (
        line.replace("[", "@<:@")
        .replace("]", "@:>@")
        .replace("#", "@%:@")
        .replace("$", "@S|@")
    )


def main(args):
    with open_input(args.input) as istr, open_output(args.output) as ostr:
        ostr.write(args.open)
        for line in istr:
            ostr.write(escape(line))
            ostr.write("\n")
        ostr.write(args.close)


if __name__ == "__main__":
    main(parse_args())



# Configuration file for the Sphinx documentation builder.

from pathlib import Path


project = "libpqxx"
copyright = "2000-2026, Jeroen T. Vermeulen"
author = "Jeroen T. Vermeulen"

version = (Path(__file__).parents[2] / "VERSION").read_text().strip()
release = ".".join(version.split(".")[:2])

extensions = [
    # TODO: Not actually using Sphinx for now.  Should we?
    # From readthedocs template:
    # 'sphinx.ext.duration',
    # 'sphinx.ext.doctest',
    # 'sphinx.ext.autodoc',
    # 'sphinx.ext.autosummary',
    # 'sphinx.ext.intersphinx',
    # Additional:
    # TODO: Not using Breathe for now.  Should we?
    # 'breathe',
    "myst_parser",
]

# TODO: What do these actually mean?
# intersphinx_mapping = {
#     'python': ('https://docs.python.org/3/', None),
#     'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
# }
# intersphinx_disabled_domains = ['std']

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"
